var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import { CryptoUtil, Uint8ArrayStack } from "./common";
import ZIP from "./zip";
/**
 * 文件名：FileSaveRegister
 * 作　者：Bryce.Li
 * 创　建：2021/6/16 11:28
 * 描　述：Bryce.Li
 * */
export default class FileSaveRegister {
    constructor(scope = '/', options = {
        confirmMessage: '',
        onMessage: (data) => {
        }
    }) {
        this.options = options;
        this.aborterMap = {};
        this.progressMap = {};
        this.completeMap = {};
        this.errorMap = {};
        this.active = false;
        // 注册service worker，并初始化逻辑
        if ('serviceWorker' in navigator) {
            navigator.serviceWorker.register('/sswa-sw.js', { scope }).then(swReg => {
                if (swReg.installing) {
                    window.location.reload();
                }
                else if (swReg.active) {
                    this.active = true;
                    return swReg;
                }
                return Promise.reject();
            }).then((reg) => {
                const messageChannel = new MessageChannel();
                reg.active.postMessage("SW_INIT", [messageChannel.port2]);
                this.messagePort = messageChannel.port1;
                // keepAlive 保存sw存活，firefox在30s后会关闭不活动的service worker
                // 通过postMessage ping pong方式使其不被关闭
                setInterval(() => reg.active.postMessage('ping'), 10000);
            }).then(() => {
                this.messagePort.onmessage = this._getMessage.bind(this);
            }).catch(err => {
            });
        }
        // 刷新或关闭页面时，清除\取消 所有下载任务
        window.addEventListener('beforeunload', () => this.abort('user refresh page'));
    }
    /**
     *  下载文件
     *  @param {string} filename 文件名
     *  @param {string | {name:string,url:string}[]} url
     *  @param {number} size 文件总大小 byte
     *  @param options
     * */
    download(filename, url, size, options = {}) {
        if (this.active) {
            const uuid = Math.random().toString().slice(-10);
            this.aborterMap[uuid] = null;
            this.messagePort.postMessage({
                type: 'DOWNLOAD_REGISTER', uuid, url, filename, size, password: options.password
            });
            options.cancel && options.cancel(() => {
                this.abort('cancel by canceler', uuid);
            });
            this.progressMap[uuid] = (loaded) => {
                options.onProgress && options.onProgress(loaded, size);
            };
            this.completeMap[uuid] = () => {
                options.onComplete && options.onComplete();
            };
            this.errorMap[uuid] = (message) => {
                options.onError && options.onError(message);
            };
        }
        else {
            console.error('service worker is not register or active');
            const bl = confirm(this.options.confirmMessage || 'sw服务未注册或激活，是否立即刷新浏览器？');
            if (bl)
                window.location.reload();
        }
    }
    /**
     * 终止下载
     * @param {string} message
     * @param {string} uuid
     * @return {boolean} 是否终止成功
     * */
    abort(message, uuid = null) {
        if (uuid) {
            if (this.aborterMap[uuid]) {
                this.aborterMap[uuid](message);
                return true;
            }
        }
        else {
            for (const uuid in this.aborterMap) {
                this.abort('user refresh page', uuid);
            }
        }
        return false;
    }
    /**
     *  MessageChannel 监听事件
     *  @param {MessageEvent} event
     * */
    _getMessage(event) {
        const { url, type, uniqLink, uuid, password } = event.data;
        this.options.onMessage && this.options.onMessage(event.data);
        if (type === 'DOWNLOAD_REGISTER_SUCCESS') {
            const iframe = document.createElement("iframe");
            iframe.src = uniqLink;
            iframe.hidden = true;
            document.body.appendChild(iframe);
            setTimeout(() => document.body.removeChild(iframe), 1000);
            if (Array.isArray(url)) {
                this._batchFetchFile(url, uniqLink, uuid, password);
            }
            else {
                this._fetchFile(url, uniqLink, uuid, password);
            }
        }
        else if (type === 'USER_CANCEL_DOWNLOAD') {
            this.abort('cancel by browser', uuid);
        }
    }
    /**
     * 连接 stream 数据流
     * @param {string} uniqLink:string
     * @param {number} uuid:number
     * @param {ReadableStream} readableStream
     * @param {string} password
     * @param zip
     * */
    _pipStream(uniqLink, uuid, readableStream, password, zip = false) {
        let loaded = 0;
        const reader = readableStream.getReader();
        this.messagePort.postMessage({ type: 'DOWNLOAD_START', uniqLink });
        const pump = () => reader.read().then(res => {
            if (res.done) {
                this.messagePort.postMessage({ type: 'DOWNLOAD_END', uniqLink });
                this.completeMap[uuid]();
            }
            else {
                const chunk = res.value;
                loaded += chunk.length;
                this.messagePort.postMessage({ type: 'DOWNLOADING', chunk, uniqLink, uuid, password, zip });
                this.progressMap[uuid](loaded);
                pump();
            }
        });
        pump();
    }
    /**
     * 增强 Fetch，具有中断及解密功能
     * _fetch
     * @param {string} url
     * @param {string} password
     * @param {string} uuid
     * @param options
     * @returns {Promise}
     * */
    _fetch(url, password, uuid, options) {
        const abort = options.canceler || function () {
        };
        let uint8ArrayStack = new Uint8ArrayStack();
        const abortHandler = (res) => {
            const reader = res.body.getReader();
            const stream = new ReadableStream({
                start: (controller) => {
                    let aborted = false, keyIV;
                    const push = () => __awaiter(this, void 0, void 0, function* () {
                        try {
                            // 增加网络异常reader.read()捕获-abort
                            const { value, done } = yield reader.read();
                            if (!done) {
                                if (!password) {
                                    controller.enqueue(value);
                                    push();
                                    return;
                                }
                                uint8ArrayStack.push(value);
                                const loopDecrypt = () => __awaiter(this, void 0, void 0, function* () {
                                    // 循环解密
                                    if (uint8ArrayStack.length > CryptoUtil.DECRYPT_BLOCK_SIZE) {
                                        const inBytes = uint8ArrayStack.readBytes(CryptoUtil.DECRYPT_BLOCK_SIZE);
                                        let { chunkPlain, nextIv } = yield CryptoUtil.chunkDecrypt(inBytes, keyIV.key, keyIV.iv);
                                        try {
                                            controller.enqueue(chunkPlain);
                                        }
                                        catch (e) {
                                        }
                                        keyIV.iv = nextIv;
                                        yield loopDecrypt();
                                    }
                                    else {
                                        push();
                                    }
                                });
                                // 获取原始key和iv
                                if (uint8ArrayStack.length >= CryptoUtil.AES_BLOCK_SIZE && !keyIV) {
                                    const inBytes = uint8ArrayStack.readBytes(CryptoUtil.AES_BLOCK_SIZE);
                                    keyIV = yield CryptoUtil.getKeyIv(inBytes, password);
                                }
                                yield loopDecrypt();
                            }
                            else {
                                if (!aborted) {
                                    if (!password) {
                                        controller.close();
                                        return;
                                    }
                                    if (uint8ArrayStack.length > 0) {
                                        // 解密剩余部分
                                        try {
                                            const plain = yield CryptoUtil.decrypt(uint8ArrayStack.readBytes(uint8ArrayStack.length), keyIV.key, keyIV.iv);
                                            controller.enqueue(new Uint8Array(plain));
                                            controller.close();
                                        }
                                        catch (e) {
                                            // 解密失败捕获-abort
                                            this.abort('decrypt error', uuid);
                                        }
                                    }
                                    else {
                                        controller.close();
                                    }
                                }
                            }
                        }
                        catch (err) {
                            this.abort(err, uuid);
                        }
                    });
                    abort(() => {
                        reader.cancel();
                        controller.error('forced interrupt fetch');
                        aborted = true;
                    });
                    push();
                }
            });
            return new Response(stream, { headers: res.headers });
        };
        return fetch(url).then(abortHandler);
    }
    /**
     * _fetchFile
     * 单文件下载
     * @param {string} url
     * @param {string} uniqLink
     * @param {string} uuid
     * @param {string} password
     * */
    _fetchFile(url, uniqLink, uuid, password) {
        this._fetch(url, password, uuid, {
            canceler: (canceler) => {
                this.aborterMap[uuid] = (message = 'abort') => {
                    canceler();
                    this.messagePort.postMessage({ uniqLink, type: 'DOWNLOAD_ABORT' });
                    this.errorMap[uuid](message);
                };
            }
        }).then(response => {
            const readableStream = response.body;
            this._pipStream(uniqLink, uuid, readableStream, password);
        }).catch((err) => {
            if (!this.abort(uuid, err)) {
                this.messagePort.postMessage({ uniqLink, type: 'DOWNLOAD_ABORT' });
                this.errorMap[uuid](err);
            }
        });
    }
    /**
     * _batchFetchFile
     * @param {{name:string,url:string}[]} fileList
     * @param {string} uniqLink
     * @param {string} uuid
     * @param password
     * */
    _batchFetchFile(fileList, uniqLink, uuid, password) {
        let fetchCanceler;
        const files = fileList.map(item => [item.path, item.url]).values();
        // @ts-ignore
        const { readableZipStream, abort } = new ZIP({
            pull: (ctrl) => {
                const it = files.next();
                if (it.done) {
                    ctrl.close();
                }
                else {
                    // TODO 这里直接捕获了异常是否影响pull
                    const [name, url] = it.value;
                    return this._fetch(url, password, uuid, {
                        canceler: canceler => fetchCanceler = canceler
                    }).then(res => {
                        ctrl.enqueue({ name, stream: () => res.body });
                    }).catch((err) => {
                        if (!this.abort(uuid, err)) {
                            this.messagePort.postMessage({ uniqLink, type: 'DOWNLOAD_ABORT' });
                            this.errorMap[uuid](err);
                        }
                    });
                }
            }
        });
        this.aborterMap[uuid] = (message = 'abort') => {
            abort();
            fetchCanceler();
            this.messagePort.postMessage({ uniqLink, type: 'DOWNLOAD_ABORT' });
            this.errorMap[uuid](message);
        };
        this._pipStream(uniqLink, uuid, readableZipStream, password, true);
    }
}
